
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <sys/time.h>

enum {
	// _xinfreq = 20_000_000,
	_xtlfreq = 20_000_000,
	_clkfreq = 300_000_000,
	DOWNLOAD_BAUD = 230_400,
	DEBUG_BAUD = DOWNLOAD_BAUD,

	MOSI = 32,
	CLK = 33,
	CS = 34,
	MISO = 35,
};


/*
 *      ____     Mode 0 : CPHA=0  CPOL=0       ____
 *   CS     |_________________________________|
 *             _   _   _   _   _   _   _   _
 *  CLK ______| |_| |_| |_| |_| |_| |_| |_| |______
 *           ___ ___ ___ ___ ___ ___ ___ ___
 * MOSI ----X___X___X___X___X___X___X___X___XXX----
 * MISO       7   6   5   4   3   2   1   0
 * 
 * 
 *      ____     Mode 3 : CPHA=1  CPOL=1       ____
 *   CS     |_________________________________|
 *      ______   _   _   _   _   _   _   _   ______
 *  CLK       |_| |_| |_| |_| |_| |_| |_| |_|
 *             ___ ___ ___ ___ ___ ___ ___ ___
 * MOSI ----XXX___X___X___X___X___X___X___X___X----
 * MISO         7   6   5   4   3   2   1   0
 * 
 * 
 * As a minimum, SPI requires just half clock cycle from CS low to first clock
 *      ____ <------>
 *   CS     |________________________ ......
 *                   _______
 *  CLK ____________|       |_______| ......
 * 
 * The easy solution is enable the serial smartpin after CS falls:
 * 		waitse1    // wait for falling edge
 * 		dirh #MOSI    // enable RX smartpin
 * However, this becomes impossible once the SPI clock is faster than the time
 * it takes to enable the smartpin.  Bear in mind that the Prop2's I/O staging
 * latency impacts how quickly software can respond to CS falling.
 * At 10 Mbit/s, there's just 50.0 nanoseconds to both identify CS low then
 * enable the smartpin before missing that first CLK edge.  This is tight for
 * software even at 200 MHz sysclock.  Whereas the smartpin can handle a bit
 * rate of up to sysclock/3.  So, 66 Mbit/s using 200 MHz sysclock.
 * 
 * Just to be clear, collecting the wrong number of clock pulses will
 * mangle your data!
 * 
 * 
 **** THE SOLUTION ****
 * 
 * The alternative is to enable the serial smartpin before CS is lowered.  This
 * allows fastest clocking but also creates a problem.  Spurious CLK pulses
 * while CS is high do happen for various reasons.  The most likely one is
 * simply because the SPI master controller has just been initailised and can
 * glitch its pins in the process.
 * 
 * So, to achieve reliable smartpin starting conditions, some crafted code is
 * needed.  There's two components, state sequence and pin detects.
 * 
 * Pin Detects:
 *    Falling CS:  This is an edge detect, it triggers once per CS cycle.
 *    Spurious CLK:  This uses a combo of CS high and CLK high together.  It's
 * important that CLK is masked when CS is low.  Without checking against CS, a
 * regular fast cycling clock would flood this event, resetting the smartpin
 * over and.over until the clock pulses cease.
 * 
 * The pattern event also performs a second critical function over what could
 * be achieved through simple fast cycling on a TESTP.  It catches the tiniest
 * clock glitches that can be missed by software sampling in a tight loop,
 * thereby accurately informing the program of smartpin's condition.
 * 
 * The pattern event uses 16 bytes of code to configure it but it surely
 * performs a vital job.
 *
 * State Sequence:
 *    The starting phase consisits of the two branch on event instructions,
 * JPAT/JNSE1, with the smarpin resetter instructions, DIRL/DIRH.  The hardware
 * event of JNSE1 wasn't strictly needed but provided a faster loop not needing
 * any TESTP.  That's worth it by itself.
 * 
 * Regarding the inner loop:  The exit state changes are decided using TESTP
 * sampling.  This turned out advantageous over using hardware events.  Both the
 * smartpin's buffer full flag and the CS pin produce a steady state indicator
 * that is good to be sampled.  This in turn allows a cleaner logic decision
 * tree.  The smartpin can raise its byte received flag two sysclock ticks after
 * CS pin rises and still be correctly emptied.
 */


static uint32_t  spi_shift_in( uint8_t *addr, size_t len )
{
    uint32_t  count = len;    // bytes received count down

    __asm volatile {  // no optimising and enforces Fcache use - Needed to free up the FIFO
		wrfast	#0, addr    // setup the hubRAM FIFO hardware for sequencial writes
		modz	15   wz    // Set Z flag for matching S operand event for SETPAT
		neg	pa, #CLK & 0x20   wc    // INA/INB port select with C flag for SETPAT
		setpat	mask1, mask1    // Setup event:  CLK high and CS high
		setse1	#2<<6 | CS    // Setup event:  1=rising-edge, 2=falling-edge, 3=both-edges, 4=low-level, 6=high-level
.retry
		dirl	#MOSI    // disable/reset SPI RX smartpin
		dirh	#MOSI    // enable SPI RX smartpin
.wait
		jpat	#.retry    // CLK toggled while CS still high, treat as a glitch
		jnse1	#.wait    // wait while CS is high

		rep	@.rend, #0    // loops forever, 14 ticks per loop
		testp	#CS   wz    // checking CS first allows final CLK close to CS rising
		testp	#MOSI   wc   // check for completed byte
	if_c	rdpin	pa, #MOSI   // collect the received byte into upper bits of PA
if_nc_and_z	jmp	#.rend    // break loop when no data and CS high
	if_c	djf	count, #.rend    // break loop when data block full
	if_c	rev	pa    // 32-bit endian bit-swap
	if_c	wfbyte	pa    // write byte to FIFO (hubRAM)
.rend
	if_nc	sub	len, count    // bytes received = bufferlength - countdown
		ret

mask1		long	(1<<(CS & 0x1f)) | (1<<(CLK & 0x1f))   // CS/CLK pin mask for SETPAT
    }  // Exiting of Fcache/Cogexec triggers an implicit RDFAST - flushing remaining FIFO content to hubRAM

    return  len;
}



static size_t  compare( void *src1, size_t size )
{
    size_t  good = 0, pattern = 0;

    __asm volatile {  // no optimising and enforces Fcache use - Needed to free up the FIFO
		rdfast  #0, src1

		rep     @.rend, size
		rfbyte  src1
		incmod  pattern, #255
		cmp     src1, pattern   wz
	if_z	add     good, #1
.rend
    }
    return good;
}



void main(void)
{
    uint32_t  m, x, ret;
    uint8_t  rxdata[40_004];

    _wrpin(CS, P_HIGH_15K | P_LOW_FLOAT | P_SYNC_IO | P_SCHMITT_A);  // do this first
    _drvh(CS);  // 15 kR pull-up resistor

    _waitms(200);
    printf( "\n   clkfreq = %d   clkmode = 0x%x\n", _clockfreq(), _clockmode() );

    _wrpin(CLK, P_SYNC_IO | P_SCHMITT_A);  // single ended
//    _wrpin(CLK, P_SYNC_IO | P_FILT0_AB | P_COMPARE_AB);  // differential pair

    m = P_SYNC_RX | P_SYNC_IO | P_SCHMITT_A;  // single ended
//    m = P_SYNC_RX | P_SYNC_IO | P_FILT0_AB | P_COMPARE_AB;  // differential pair
    m |= ((CLK-MOSI) & 0b111) << 24;  // smartB (clock pin) input select
    x = 0b1_00000 | (8-1);  // late sample alignment and 8-bit words
    _pinstart(MOSI, m, x, 0);


    while(1)
    {
        ret = spi_shift_in( rxdata, sizeof(rxdata) );
        printf( " Byte Count = %d", ret );
        if( ret >= 8 )
            printf( "    Mismatch = %d    BlockID = %d", ret - 1 - compare( &rxdata[1], ret - 1 ), rxdata[0] );
        printf( "\n" );
/*
        for(unsigned i = 0; i < ret; i++)
            printf(" %02x",rxdata[i]);
        printf( "\n" );
*/
    }
}
